#ifndef LIBSYS_H
#define LIBSYS_H

// ---------------------------------------------------------
// Types
// ---------------------------------------------------------
#define INT8		signed char
#define INT16		signed short
#define INT32		signed int
#define INT64		signed long long
#define UINT8		unsigned char
#define UINT16		unsigned short
#define UINT32		unsigned int
#define UINT64		unsigned long long
#define INT			INT32
#define UINT		UINT32
#define FLOAT32		float
#define FLOAT64		double


#define	sys_gpio0		0xA0000000
#define	sys_gpio1		0xA0000004
#define	sys_led_port	sys_gpio0
#define	sys_usb_ctrl	sys_gpio1
#define	sys_timer_usec	0xA0000008
#define	sys_timer_sec	0xA000000C
#define	sys_uart_data	0xA0010000
#define	sys_uart_stat	0xA0010004
#define	sys_uart_baud	0xA0010008
#define	sys_usb_data	0xA0020000
#define	sys_usb_mbx		0xA0020004
#define	sys_usb_addr	0xA0020008
#define	sys_usb_status	0xA002000C
#define	sys_vga_data	0xA0030004
#define	sys_vga_posx	0xA0030008
#define	sys_vga_posy	0xA0030010
#define	sys_vga_clrscr	0xA0030020
#define	sys_vga_clrline	0xA0030040
#define	sys_vga_mctrl	0xA0030100
#define	sys_vga_moffs	0xA0030200
#define	sys_ac97_stat	0xA0040000
#define	sys_ac97_ctrl	0xA0040000
#define	sys_ac97_acstat	0xA0040400
#define	sys_ac97_acctrl	0xA0040600
#define	sys_ac97_pcm	0xA0040800
#define	sys_ac97_wavin	sys_ac97_pcm
#define	sys_ac97_wavout	sys_ac97_pcm
#define	sys_flash_io	0xA4000000
#define	sys_ssram_io	0xA8000000

//#define STDOUT_FUNCTION	_cg_putchar		// Video character
//#define STDERR_FUNCTION	_putchar		// Serial output

#ifndef STDOUT_FUNCTION
#define STDOUT_FUNCTION	_putchar		// Serial output
#endif
#ifndef STDERR_FUNCTION
#define STDERR_FUNCTION	_putchar		// Serial output
#endif

UINT32 CP0_SR_read(void);
void CP0_SR_write(UINT32 val);
UINT32 CP0_CR_read(void);
void CP0_CR_write(UINT32 val);
UINT32 CP0_PRID_read(void);
UINT32 CP0_TR_read(void);
void CP0_TR_write(UINT32 val);

char readchar(void);
void writechar(char c);
int write(int file, char *ptr, int len);
int sputs(char *pStr);
void print_byte(char byte);
void print_word(int word);
void _exit (int exitcode);
void sleep(unsigned ms);

void PrintBuffer8(UINT8 *pBuf, int nbpr, int len);
void memdump(UINT8 *pBuf, int print_offset_only, int num_bytes_per_row, int len);

#endif // LIBSYS_H

